﻿Imports System
Imports System.IO.Ports
Imports System.Threading
Imports System.Media

Public Class Form_24084R

    Const STR_MODEL As String = " PM-24084R "                   ' Model: PM-24084R

    ' Command ******************************************************************
    Const CMD_READ_DATA As Byte = &H80                          ' Command: Read Data (讀取資料)
    Const CMD_SYNC_EN As Byte = &H70                            ' Command: SYNC Enable (同步致能)
    Const CMD_L1_SET As Byte = &H91                             ' Command: L1 Set (L1設定)
    Const CMD_L2_SET As Byte = &H92                             ' Command: L2 Set (L2設定)
    Const CMD_L3_SET As Byte = &H93                             ' Command: L3 Set (L3設定)
    Const CMD_L4_SET As Byte = &H94                             ' Command: L4 Set (L4設定)
    Const CMD_L1_OUTPUT As Byte = &HA1                          ' Command: L1 Output (L1輸出電流)
    Const CMD_L2_OUTPUT As Byte = &HA2                          ' Command: L2 Output (L2輸出電流)
    Const CMD_L3_OUTPUT As Byte = &HA3                          ' Command: L3 Output (L3輸出電流)
    Const CMD_L4_OUTPUT As Byte = &HA4                          ' Command: L4 Output (L4輸出電流)
    Const CMD_L1_ON_OFF As Byte = &HC1                          ' Command: L1 ON/OFF
    Const CMD_L2_ON_OFF As Byte = &HC2                          ' Command: L2 ON/OFF
    Const CMD_L3_ON_OFF As Byte = &HC3                          ' Command: L3 ON/OFF
    Const CMD_L4_ON_OFF As Byte = &HC4                          ' Command: L4 ON/OFF

    ' ACK_State ================================================================
    Const AS_FAILED As Byte = &H80                              ' bit7 命令執行狀態    : 0=成功, 1=失敗
    Const AS_SYNC_EN1 As Byte = &H40                            ' bit6~5 ON/OFF同步致能: 00=ASYNC (全部個別控制)
    Const AS_SYNC_EN0 As Byte = &H20                            '                        01=SYNC2 (L1~L2同步控制)
    '                                                                                    10=SYNC3 (L1~L3同步控制)
    '                                                                                    11=SYNC4 (L1~L4同步控制)
    Const AS_PWR_ERR As Byte = &H10                             ' bit4 電源狀態        : 0=OK, 1=Error
    Const AS_L4_ERR As Byte = &H8                               ' bit3 L4狀態          : 0=OK, 1=Error
    Const AS_L3_ERR As Byte = &H4                               ' bit2 L3狀態          : 0=OK, 1=Error
    Const AS_L2_ERR As Byte = &H2                               ' bit1 L2狀態          : 0=OK, 1=Error
    Const AS_L1_ERR As Byte = &H1                               ' bit0 L1狀態          : 0=OK, 1=Error

    ' SYNC_EN ==================================================================
    Const ASYNC As Byte = 0                                     ' ASYNC (全部個別控制)
    Const SYNC2 As Byte = 1                                     ' SYNC2 (L1~L2同步控制)
    Const SYNC3 As Byte = 2                                     ' SYNC3 (L1~L3同步控制)
    Const SYNC4 As Byte = 3                                     ' SYNC4 (L1~L4同步控制)

    ' Lx_Set ===================================================================
    Const SET_EXT_CTRL As Byte = &H2                            ' bit1 外部ON/OFF控制: 0=Disable, 1=Enable
    Const SET_ON_OFF As Byte = &H1                              ' bit0 ON/OFF        : 0=OFF, 1=ON

    ' ==========================================================================
    Const Lines_Max As UShort = 256                             ' 最大行數
    Const ACK_TO As Byte = 10                                   ' 回應逾時, 約500ms (10 * 50ms)
    Const AUTO_TMR As Byte = 60                                 ' 自動詢問計時, 約3sec (60 * 50ms)

    ' **************************************************************************
    Dim Line_No As UShort = 0                                   ' 行數編號
    Dim Line_CTR As UShort = 0                                  ' 行數計數
    Dim PortNo() As String                                      ' 埠編號
    Dim PORT_Flag As Boolean = False                            ' PORT 旗標
    Dim READ_Flag As Boolean = False                            ' READ 旗標
    Dim ACK_TO_CTR As Byte = 0                                  ' 回應逾時計時
    Dim ACK_State As Byte = 0                                   ' 回應狀態
    Dim Auto_CTR As Byte = 0                                    ' 自動詢問計數
    Dim Flash_CTR As Byte = 0                                   ' 閃爍計數
    Dim MaxCurrent() As Integer = {500, 600, 700, 800}

    ' TX =======================================================================
    Dim TX_Flag As Boolean = False                              ' TX 旗標
    Dim TX_Packet(7) As Byte                                    ' TX 封包
    Dim TX_CMD As Byte = 0                                      ' TX 命令
    Dim TX_CTR As Byte = 0                                      ' TX 傳送次數

    ' RX =======================================================================
    Dim RX_Flag As Boolean = False                              ' RX 旗標
    Dim RX_Buffer(128) As Byte                                  ' RX 緩衝區
    Dim RX_Packet(32) As Byte                                   ' RX 封包
    Dim RX_Index As Byte = 0                                    ' RX 封包索引
    Dim RX_Length As Byte = 0                                   ' RX 封包長度

    ' ==========================================================================
    Dim SYNC_EN_CHG As Boolean = False                          ' SYNC Enable 已改變
    Dim SYNC_EN As Byte = ASYNC                                 ' SYNC Enable

    ' L1 =======================================================================
    Dim L1_Set_CHG As Boolean = False                           ' L1 Set 已改變
    Dim L1_Output_CHG As Boolean = False                        ' L1 Output 已改變
    Dim L1_ON_OFF_CHG As Boolean = False                        ' L1 ON/OFF 已改變
    Dim L1_Set As Byte = 0                                      ' L1 設定值
    Dim L1_Output As Integer = 0                                ' L1 輸出值

    ' L2 =======================================================================
    Dim L2_Set_CHG As Boolean = False                           ' L2 Set 已改變
    Dim L2_Output_CHG As Boolean = False                        ' L2 Output 已改變
    Dim L2_ON_OFF_CHG As Boolean = False                        ' L2 ON/OFF 已改變
    Dim L2_Set As Byte = 0                                      ' L2 設定值
    Dim L2_Output As Integer = 0                                ' L2 輸出值

    ' L3 =======================================================================
    Dim L3_Set_CHG As Boolean = False                           ' L3 Set 已改變
    Dim L3_Output_CHG As Boolean = False                        ' L3 Output 已改變
    Dim L3_ON_OFF_CHG As Boolean = False                        ' L3 ON/OFF 已改變
    Dim L3_Set As Byte = 0                                      ' L3 設定值
    Dim L3_Output As Integer = 0                                ' L3 輸出值

    ' L4 =======================================================================
    Dim L4_Set_CHG As Boolean = False                           ' L4 Set 已改變
    Dim L4_Output_CHG As Boolean = False                        ' L4 Output 已改變
    Dim L4_ON_OFF_CHG As Boolean = False                        ' L4 ON/OFF 已改變
    Dim L4_Set As Byte = 0                                      ' L4 設定值
    Dim L4_Output As Integer = 0                                ' L4 輸出值

    ' **************************************************************************

    Private Sub Form_24084R_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        cbb_L1_MaxC.SelectedIndex = 0                           ' 500mA
        cbb_L2_MaxC.SelectedIndex = 0                           ' 500mA
        cbb_L3_MaxC.SelectedIndex = 0                           ' 500mA
        cbb_L4_MaxC.SelectedIndex = 0                           ' 500mA
        COMPort_Check()                                         ' 串列埠檢知
    End Sub

    Private Sub timer_Tick(sender As Object, e As EventArgs) Handles timer.Tick
        If PORT_Flag = True Then                                ' 串列埠已開啟?
            If RX_Flag = True Then                              ' 已接收封包?
                RX_Flag = False
                TX_Flag = False
                TX_CTR = 0
                ACK_TO_CTR = 0
                If READ_Flag = False Then
                    Update_Lx()                                 ' 更新 (Lx參數)
                    Update_ACK()                                ' 更新 (ACK)
                    READ_Flag = True
                Else
                    Update_ACK()                                ' 更新 (ACK)
                End If
            Else
                If TX_Flag = True Then                          ' 已傳送命令?
                    Check_RX_TO()                               ' 檢查接收逾時
                Else
                    Check_TX_CMD()                              ' 檢查傳送命令
                End If
                Auto_Ask()                                      ' 自動詢問
            End If
        End If
        ERR_Flash()                                             ' 錯誤提示閃爍
    End Sub

    ' ==========================================================================

    Private Sub Check_RX_TO()                                   ' 檢查接收逾時
        ACK_TO_CTR += 1
        If ACK_TO_CTR >= ACK_TO Then
            ACK_TO_CTR = 0
            If TX_CTR > 2 Then
                Dsp_String(Color.Red, STR_MODEL + "逾時未回應 !" + vbCrLf)
                COMPort_Disconnect()                            ' 串列埠離線
                SystemSounds.Beep.Play()
            Else
                Select Case TX_CMD
                    Case CMD_READ_DATA
                        TX_Read_Data()                          ' 傳送: Read Data
                    Case CMD_SYNC_EN
                        TX_SYNC_EN()                            ' 傳送: SYNC Enable
                    Case CMD_L1_SET
                        TX_L1_Set()                             ' 傳送: L1 Set
                    Case CMD_L2_SET
                        TX_L2_Set()                             ' 傳送: L2 Set
                    Case CMD_L3_SET
                        TX_L3_Set()                             ' 傳送: L3 Set
                    Case CMD_L4_SET
                        TX_L4_Set()                             ' 傳送: L4 Set
                    Case CMD_L1_OUTPUT
                        TX_L1_Output()                          ' 傳送: L1 Output
                    Case CMD_L2_OUTPUT
                        TX_L2_Output()                          ' 傳送: L2 Output
                    Case CMD_L3_OUTPUT
                        TX_L3_Output()                          ' 傳送: L3 Output
                    Case CMD_L4_OUTPUT
                        TX_L4_Output()                          ' 傳送: L4 Output
                    Case CMD_L1_ON_OFF
                        TX_L1_ON_OFF()                          ' 傳送: L1 ON/OFF
                    Case CMD_L2_ON_OFF
                        TX_L2_ON_OFF()                          ' 傳送: L2 ON/OFF
                    Case CMD_L3_ON_OFF
                        TX_L3_ON_OFF()                          ' 傳送: L3 ON/OFF
                    Case CMD_L4_ON_OFF
                        TX_L4_ON_OFF()                          ' 傳送: L4 ON/OFF
                End Select
            End If
        End If
    End Sub

    Private Sub Check_TX_CMD()                                  ' 檢查傳送命令
        If SYNC_EN_CHG = True Then
            TX_SYNC_EN()                                        ' 傳送: SYNC Enable
            Return
        End If

        If L1_Set_CHG = True Then
            TX_L1_Set()                                         ' 傳送: L1 Set
            Return
        End If

        If L2_Set_CHG = True Then
            TX_L2_Set()                                         ' 傳送: L2 Set
            Return
        End If

        If L3_Set_CHG = True Then
            TX_L3_Set()                                         ' 傳送: L3 Set
            Return
        End If

        If L4_Set_CHG = True Then
            TX_L4_Set()                                         ' 傳送: L4 Set
            Return
        End If

        If L1_Output_CHG = True Then
            TX_L1_Output()                                      ' 傳送: L1 Output
            Return
        End If

        If L2_Output_CHG = True Then
            TX_L2_Output()                                      ' 傳送: L2 Output
            Return
        End If

        If L3_Output_CHG = True Then
            TX_L3_Output()                                      ' 傳送: L3 Output
            Return
        End If

        If L4_Output_CHG = True Then
            TX_L4_Output()                                      ' 傳送: L4 Output
            Return
        End If

        If L1_ON_OFF_CHG = True Then
            TX_L1_ON_OFF()                                      ' 傳送: L1 ON/OFF
            Return
        End If

        If L2_ON_OFF_CHG = True Then
            TX_L2_ON_OFF()                                      ' 傳送: L2 ON/OFF
            Return
        End If

        If L3_ON_OFF_CHG = True Then
            TX_L3_ON_OFF()                                      ' 傳送: L3 ON/OFF
            Return
        End If

        If L4_ON_OFF_CHG = True Then
            TX_L4_ON_OFF()                                      ' 傳送: L4 ON/OFF
            Return
        End If
    End Sub

    Private Sub Auto_Ask()                                      ' 自動詢問
        Auto_CTR += 1
        If Auto_CTR >= AUTO_TMR Then
            TX_Read_Data()                                      ' 傳送: Read Data
        End If

        If cbb_L1_MaxC.DroppedDown = True Then
            Auto_CTR = 0
        End If

        If cbb_L2_MaxC.DroppedDown = True Then
            Auto_CTR = 0
        End If

        If cbb_L3_MaxC.DroppedDown = True Then
            Auto_CTR = 0
        End If

        If cbb_L4_MaxC.DroppedDown = True Then
            Auto_CTR = 0
        End If

        If nud_L1_Output.Focused = True Then
            Auto_CTR = 0
        End If

        If nud_L2_Output.Focused = True Then
            Auto_CTR = 0
        End If

        If nud_L3_Output.Focused = True Then
            Auto_CTR = 0
        End If

        If nud_L4_Output.Focused = True Then
            Auto_CTR = 0
        End If

        If cbb_SYNC_EN.DroppedDown = True Then
            Auto_CTR = 0
        End If
    End Sub

    Private Sub ERR_Flash()                                     ' 錯誤提示閃爍
        If Flash_CTR = 0 Then
            If (ACK_State And AS_PWR_ERR) = AS_PWR_ERR Then
                lb_PWR_ERR.BackColor = Color.Red
            End If
            If (ACK_State And AS_L4_ERR) = AS_L4_ERR Then
                lb_L4.BackColor = Color.Red
            End If
            If (ACK_State And AS_L3_ERR) = AS_L3_ERR Then
                lb_L3.BackColor = Color.Red
            End If
            If (ACK_State And AS_L2_ERR) = AS_L2_ERR Then
                lb_L2.BackColor = Color.Red
            End If
            If (ACK_State And AS_L1_ERR) = AS_L1_ERR Then
                lb_L1.BackColor = Color.Red
            End If
        End If

        If Flash_CTR = 5 Then
            If (ACK_State And AS_PWR_ERR) = AS_PWR_ERR Then
                lb_PWR_ERR.BackColor = Color.Black
            End If
            If (ACK_State And AS_L4_ERR) = AS_L4_ERR Then
                lb_L4.BackColor = Color.Black
            End If
            If (ACK_State And AS_L3_ERR) = AS_L3_ERR Then
                lb_L3.BackColor = Color.Black
            End If
            If (ACK_State And AS_L2_ERR) = AS_L2_ERR Then
                lb_L2.BackColor = Color.Black
            End If
            If (ACK_State And AS_L1_ERR) = AS_L1_ERR Then
                lb_L1.BackColor = Color.Black
            End If
        End If

        Flash_CTR += 1
        If Flash_CTR > 9 Then
            Flash_CTR = 0
        End If
    End Sub

    ' ==========================================================================

    Private Sub Update_Lx()                                     ' 更新 (Lx參數)
        Dsp_RX_Packet(SystemColors.WindowText, 17)

        Dim v As Byte = 0
        If (RX_Packet(3) And AS_SYNC_EN1) = AS_SYNC_EN1 Then
            v = v Or 2
        End If
        If (RX_Packet(3) And AS_SYNC_EN0) = AS_SYNC_EN0 Then
            v = v Or 1
        End If
        SYNC_EN = v
        cbb_SYNC_EN.SelectedIndex = v

        Update_L1()                                             ' 更新 (L1)
        Update_L2()                                             ' 更新 (L2)
        Update_L3()                                             ' 更新 (L3)
        Update_L4()                                             ' 更新 (L4)
    End Sub

    Private Sub Update_L1()                                     ' 更新 (L1)
        L1_Set = RX_Packet(4) And &H33
        If (L1_Set And SET_ON_OFF) = 0 Then                     ' L1 ON/OFF ?
            pn_L1_LED.BackColor = Color.Black
            btn_L1.Text = "ON"
        Else
            pn_L1_LED.BackColor = Color.Green
            btn_L1.Text = "OFF"
        End If
        If (L1_Set And SET_EXT_CTRL) = 0 Then                   ' L1外部ON/OFF控制 ?
            ckb_L1_EXT.Checked = False
            pn_L1_LED.Visible = True
            btn_L1.Visible = True
        Else
            ckb_L1_EXT.Checked = True
            pn_L1_LED.Visible = False
            btn_L1.Visible = False
        End If

        Dim MaxC, Output As Integer
        cbb_L1_MaxC.SelectedIndex = (L1_Set >> 4) And &H3
        MaxC = MaxCurrent(cbb_L1_MaxC.SelectedIndex)
        Output = RX_Packet(5)
        Output <<= 8
        Output = Output Or RX_Packet(6)
        nud_L1_Output.Maximum = MaxC
        tkb_L1_Output.Maximum = MaxC
        nud_L1_Output.Value = Output
        tkb_L1_Output.Value = Output
        tkb_L1_Output.TickFrequency = MaxC / 10
        lb_L1_PCV.Text = String.Format("{0:F1}", (Output * 100) / MaxC)
        L1_Output = Output
    End Sub

    Private Sub Update_L2()                                     ' 更新 (L2)
        L2_Set = RX_Packet(7) And &H33
        If (L2_Set And SET_ON_OFF) = 0 Then                     ' L2 ON/OFF ?
            pn_L2_LED.BackColor = Color.Black
            btn_L2.Text = "ON"
        Else
            pn_L2_LED.BackColor = Color.Green
            btn_L2.Text = "OFF"
        End If
        If (L2_Set And SET_EXT_CTRL) = 0 Then                   ' L2外部ON/OFF控制 ?
            ckb_L2_EXT.Checked = False
            pn_L2_LED.Visible = True
            btn_L2.Visible = True
        Else
            ckb_L2_EXT.Checked = True
            pn_L2_LED.Visible = False
            btn_L2.Visible = False
        End If

        Dim MaxC, Output As Integer
        cbb_L2_MaxC.SelectedIndex = (L2_Set >> 4) And &H3
        MaxC = MaxCurrent(cbb_L2_MaxC.SelectedIndex)
        Output = RX_Packet(8)
        Output <<= 8
        Output = Output Or RX_Packet(9)
        nud_L2_Output.Maximum = MaxC
        tkb_L2_Output.Maximum = MaxC
        nud_L2_Output.Value = Output
        tkb_L2_Output.Value = Output
        tkb_L2_Output.TickFrequency = MaxC / 10
        lb_L2_PCV.Text = String.Format("{0:F1}", (Output * 100) / MaxC)
        L2_Output = Output
    End Sub

    Private Sub Update_L3()                                     ' 更新 (L3)
        L3_Set = RX_Packet(10) And &H33
        If (L3_Set And SET_ON_OFF) = 0 Then                     ' L3 ON/OFF ?
            pn_L3_LED.BackColor = Color.Black
            btn_L3.Text = "ON"
        Else
            pn_L3_LED.BackColor = Color.Green
            btn_L3.Text = "OFF"
        End If
        If (L3_Set And SET_EXT_CTRL) = 0 Then                   ' L3外部ON/OFF控制 ?
            ckb_L3_EXT.Checked = False
            pn_L3_LED.Visible = True
            btn_L3.Visible = True
        Else
            ckb_L3_EXT.Checked = True
            pn_L3_LED.Visible = False
            btn_L3.Visible = False
        End If

        Dim MaxC, Output As Integer
        cbb_L3_MaxC.SelectedIndex = (L3_Set >> 4) And &H3
        MaxC = MaxCurrent(cbb_L3_MaxC.SelectedIndex)
        Output = RX_Packet(11)
        Output <<= 8
        Output = Output Or RX_Packet(12)
        nud_L3_Output.Maximum = MaxC
        tkb_L3_Output.Maximum = MaxC
        nud_L3_Output.Value = Output
        tkb_L3_Output.Value = Output
        tkb_L3_Output.TickFrequency = MaxC / 10
        lb_L3_PCV.Text = String.Format("{0:F1}", (Output * 100) / MaxC)
        L3_Output = Output
    End Sub

    Private Sub Update_L4()                                     ' 更新 (L4)
        L4_Set = RX_Packet(13) And &H33
        If (L4_Set And SET_ON_OFF) = 0 Then                     ' L4 ON/OFF ?
            pn_L4_LED.BackColor = Color.Black
            btn_L4.Text = "ON"
        Else
            pn_L4_LED.BackColor = Color.Green
            btn_L4.Text = "OFF"
        End If
        If (L4_Set And SET_EXT_CTRL) = 0 Then                   ' L4外部ON/OFF控制 ?
            ckb_L4_EXT.Checked = False
            pn_L4_LED.Visible = True
            btn_L4.Visible = True
        Else
            ckb_L4_EXT.Checked = True
            pn_L4_LED.Visible = False
            btn_L4.Visible = False
        End If

        Dim MaxC, Output As Integer
        cbb_L4_MaxC.SelectedIndex = (L4_Set >> 4) And &H3
        MaxC = MaxCurrent(cbb_L4_MaxC.SelectedIndex)
        Output = RX_Packet(14)
        Output <<= 8
        Output = Output Or RX_Packet(15)
        nud_L4_Output.Maximum = MaxC
        tkb_L4_Output.Maximum = MaxC
        nud_L4_Output.Value = Output
        tkb_L4_Output.Value = Output
        tkb_L4_Output.TickFrequency = MaxC / 10
        lb_L4_PCV.Text = String.Format("{0:F1}", (Output * 100) / MaxC)
        L4_Output = Output
    End Sub

    Private Sub Update_ACK()                                    ' 更新 (ACK)
        If RX_Length = 5 Then
            Dsp_RX_Packet(SystemColors.WindowText, 5)
        End If

        ACK_State = RX_Packet(3)
        If (ACK_State And &H9F) <> 0 Then                       ' AS_FAILED Or AS_PWR_ERR Or AS_L4_ERR Or AS_L3_ERR Or AS_L2_ERR Or AS_L1_ERR
            SystemSounds.Beep.Play()
        End If

        If (ACK_State And AS_FAILED) = AS_FAILED Then
            Dsp_String(Color.Red, STR_MODEL + "命令執行失敗 !" + vbCrLf)
        End If

        If (ACK_State And AS_L1_ERR) = AS_L1_ERR Then
            lb_L1.Text = "L1 Error"
            Dsp_String(Color.Red, STR_MODEL + "L1錯誤 !" + vbCrLf)
        Else
            lb_L1.Text = "L1"
            lb_L1.BackColor = Color.Black
        End If

        If (ACK_State And AS_L2_ERR) = AS_L2_ERR Then
            lb_L2.Text = "L2 Error"
            Dsp_String(Color.Red, STR_MODEL + "L2錯誤 !" + vbCrLf)
        Else
            lb_L2.Text = "L2"
            lb_L2.BackColor = Color.Black
        End If

        If (ACK_State And AS_L3_ERR) = AS_L3_ERR Then
            lb_L3.Text = "L3 Error"
            Dsp_String(Color.Red, STR_MODEL + "L3錯誤 !" + vbCrLf)
        Else
            lb_L3.Text = "L3"
            lb_L3.BackColor = Color.Black
        End If

        If (ACK_State And AS_L4_ERR) = AS_L4_ERR Then
            lb_L4.Text = "L4 Error"
            Dsp_String(Color.Red, STR_MODEL + "L4錯誤 !" + vbCrLf)
        Else
            lb_L4.Text = "L4"
            lb_L4.BackColor = Color.Black
        End If

        If (ACK_State And AS_PWR_ERR) = AS_PWR_ERR Then
            lb_PWR_ERR.ForeColor = Color.White
            Dsp_String(Color.Red, STR_MODEL + "電源過低 !" + vbCrLf)
            gpb_ON_OFF.Enabled = False
            gpb_L1.Enabled = False
            gpb_L2.Enabled = False
            gpb_L3.Enabled = False
            gpb_L4.Enabled = False
        Else
            lb_PWR_ERR.ForeColor = Color.Gray
            lb_PWR_ERR.BackColor = Color.Black
            gpb_ON_OFF.Enabled = True
            If (ACK_State And AS_L1_ERR) = 0 Then
                gpb_L1.Enabled = True
            Else
                gpb_L1.Enabled = False
            End If
            If (ACK_State And AS_L2_ERR) = 0 Then
                gpb_L2.Enabled = True
            Else
                gpb_L2.Enabled = False
            End If
            If (ACK_State And AS_L3_ERR) = 0 Then
                gpb_L3.Enabled = True
            Else
                gpb_L3.Enabled = False
            End If
            If (ACK_State And AS_L4_ERR) = 0 Then
                gpb_L4.Enabled = True
            Else
                gpb_L4.Enabled = False
            End If
        End If
    End Sub

    ' RichTextBox 顯示 *********************************************************

    Private Sub Dsp_String(ByVal text_color As Color, ByVal str As String)  ' 顯示字串
        Line_CTR = Line_CTR + 1
        If Line_CTR > Lines_Max Then
            Line_CTR = 0
            richTextBox.Clear()
        End If

        richTextBox.SelectionStart = richTextBox.Text.Length
        richTextBox.ScrollToCaret()
        richTextBox.SelectionBackColor = Color.Gainsboro
        richTextBox.SelectionColor = SystemColors.WindowText
        richTextBox.AppendText(String.Format("{0:X4}", Line_No))
        richTextBox.SelectionBackColor = SystemColors.Window
        richTextBox.SelectionColor = text_color
        richTextBox.AppendText(str)
        richTextBox.Select()

        If Line_No < 65535 Then
            Line_No += 1
        Else
            Line_No = 0
        End If
    End Sub

    Private Sub Dsp_TX_Packet(ByVal text_color As Color, ByVal length As Byte, ByVal Counter As Byte)  ' 顯示 TX 封包
        Line_CTR = Line_CTR + 1
        If Line_CTR > Lines_Max Then
            Line_CTR = 0
            richTextBox.Clear()
        End If

        richTextBox.SelectionStart = richTextBox.Text.Length
        richTextBox.ScrollToCaret()
        richTextBox.SelectionBackColor = Color.Gainsboro
        richTextBox.SelectionColor = SystemColors.WindowText
        richTextBox.AppendText(String.Format("{0:X4}", Line_No))
        richTextBox.SelectionBackColor = SystemColors.Window
        richTextBox.SelectionColor = text_color

        Dim str As String = " TX:"
        Dim i As Byte
        For i = 0 To (length - 1)
            str += String.Format(" {0:X2}", TX_Packet(i))
        Next
        str += " (" + Convert.ToString(Counter) + ")" + vbCrLf
        richTextBox.AppendText(str)
        richTextBox.Select()

        If Line_No < 65535 Then
            Line_No += 1
        Else
            Line_No = 0
        End If
    End Sub

    Private Sub Dsp_RX_Packet(ByVal text_color As Color, ByVal length As Byte)  ' 顯示 RX 封包
        Line_CTR = Line_CTR + 1
        If Line_CTR > Lines_Max Then
            Line_CTR = 0
            richTextBox.Clear()
        End If

        richTextBox.SelectionStart = richTextBox.Text.Length
        richTextBox.ScrollToCaret()
        richTextBox.SelectionBackColor = Color.Gainsboro
        richTextBox.SelectionColor = SystemColors.WindowText
        richTextBox.AppendText(String.Format("{0:X4}", Line_No))
        richTextBox.SelectionBackColor = SystemColors.Window
        richTextBox.SelectionColor = text_color

        Dim str As String = " RX:"
        Dim i As Byte
        For i = 0 To (length - 1)
            str += String.Format(" {0:X2}", RX_Packet(i))
        Next
        str += vbCrLf
        richTextBox.AppendText(str)
        richTextBox.Select()

        If Line_No < 65535 Then
            Line_No += 1
        Else
            Line_No = 0
        End If
    End Sub

    ' 串列埠 *******************************************************************

    Private Sub cbb_PortNo_Click(sender As Object, e As EventArgs) Handles cbb_PortNo.Click
        COMPort_Check()                                         ' 串列埠檢知
    End Sub

    Private Sub btn_Connect_Click(sender As Object, e As EventArgs) Handles btn_Connect.Click
        Dim state As Boolean = False

        If serialPort.IsOpen Then                               ' 串列埠是否已開啟?
            COMPort_Disconnect()                                ' 串列埠離線
        Else
            PortNo = SerialPort.GetPortNames()                  ' 取得存在的 COM Port.
            If PortNo.Length >= 1 Then
                If cbb_PortNo.Text = "" Then
                    Dsp_String(Color.Red, " 未選擇串列埠 !" + vbCrLf)
                    SystemSounds.Beep.Play()
                Else
                    For Each port As String In PortNo
                        If port = cbb_PortNo.Text Then
                            state = True
                            Exit For
                        End If
                    Next port

                    If state = True Then
                        If serialPort.IsOpen Then
                            COMPort_Disconnect()                ' 串列埠離線
                        Else
                            COMPort_Connect()                   ' 串列埠連線
                            If PORT_Flag = True Then
                                TX_Read_Data()                  ' 傳送: Read Data
                            End If
                        End If
                    End If
                End If
            Else
                cbb_PortNo.Items.Clear()                        ' 清除 ComboBox.Items 的內容.
                Dsp_String(Color.Red, " 未發現任何串列埠 !" + vbCrLf)
                SystemSounds.Beep.Play()
            End If
        End If
    End Sub

    Private Sub COMPort_Check()                                 ' 串列埠檢知
        Dim Port_No As String = cbb_PortNo.Text

        PortNo = SerialPort.GetPortNames()                      ' 取得存在的 COM Port.
        cbb_PortNo.Items.Clear()                                ' 清除 ComboBox.Items 的內容.
        If PortNo.Length >= 1 Then
            For Each port As String In PortNo                   ' 將找到的現有 COM Port 加入 ComboBox.Items
                cbb_PortNo.Items.Add(port)
                If port = Port_No Then
                    cbb_PortNo.Text = Port_No
                End If
            Next port

            If cbb_PortNo.Text = "" Then
                cbb_PortNo.Text = PortNo(0)                     ' ComboBox.Text 先顯示一個現存的 COM Port.
            End If
        Else
            Dsp_String(Color.Red, " 未發現任何串列埠 !" + vbCrLf)
            SystemSounds.Beep.Play()
        End If
    End Sub

    Private Sub COMPort_Connect()                               ' 串列埠連線
        PORT_Flag = False
        RX_Flag = False
        RX_Index = 0
        TX_Flag = False
        TX_CTR = 0
        TX_CMD = 0
        SYNC_EN_CHG = False
        L1_Set_CHG = False
        L1_Output_CHG = False
        L1_ON_OFF_CHG = False
        L2_Set_CHG = False
        L2_Output_CHG = False
        L2_ON_OFF_CHG = False
        L3_Set_CHG = False
        L3_Output_CHG = False
        L3_ON_OFF_CHG = False
        L4_Set_CHG = False
        L4_Output_CHG = False
        L4_ON_OFF_CHG = False
        serialPort.PortName = cbb_PortNo.Text

        Try
            serialPort.Open()                                   ' 開啟新的序列埠連線.
            serialPort.DiscardInBuffer()                        ' 捨棄序列驅動程式接收緩衝區的資料.
            serialPort.DiscardOutBuffer()                       ' 捨棄序列驅動程式傳輸緩衝區的資料.
            cbb_PortNo.Enabled = False
            btn_Connect.Text = "Disconnect"
            tss_Status.Text = "Connect"
            Dsp_String(Color.Magenta, " " + cbb_PortNo.Text + " connect." + vbCrLf)
            PORT_Flag = True
        Catch ex As Exception
            Dsp_String(Color.Red, " 開啟串列埠失敗 !" + vbCrLf)
            SystemSounds.Beep.Play()
        End Try
    End Sub

    Private Sub COMPort_Disconnect()                            ' 串列埠離線
        PORT_Flag = False
        RX_Flag = False
        RX_Index = 0
        TX_Flag = False
        TX_CTR = 0
        TX_CMD = 0
        ACK_State = 0
        SYNC_EN_CHG = False
        L1_Set_CHG = False
        L1_Output_CHG = False
        L1_ON_OFF_CHG = False
        L2_Set_CHG = False
        L2_Output_CHG = False
        L2_ON_OFF_CHG = False
        L3_Set_CHG = False
        L3_Output_CHG = False
        L3_ON_OFF_CHG = False
        L4_Set_CHG = False
        L4_Output_CHG = False
        L4_ON_OFF_CHG = False

        Try
            serialPort.Close()                                  ' 關閉連接埠連線.
            Dsp_String(Color.Magenta, " " + cbb_PortNo.Text + " disconnect." + vbCrLf)
        Catch ex As Exception
            COMPort_Check()                                     ' 串列埠檢知
        End Try

        cbb_PortNo.Enabled = True
        btn_Connect.Text = "Connect"
        tss_Status.Text = "Disconnect"
        lb_L1.Text = "L1"
        lb_L2.Text = "L2"
        lb_L3.Text = "L3"
        lb_L4.Text = "L4"
        lb_L1.BackColor = Color.Black
        lb_L2.BackColor = Color.Black
        lb_L3.BackColor = Color.Black
        lb_L4.BackColor = Color.Black
        lb_PWR_ERR.ForeColor = Color.Gray
        lb_PWR_ERR.BackColor = Color.Black
        gpb_ON_OFF.Enabled = False
        gpb_L1.Enabled = False
        gpb_L2.Enabled = False
        gpb_L3.Enabled = False
        gpb_L4.Enabled = False
    End Sub

    ' 串列埠傳送 ********************************************************************

    Private Sub TX_Read_Data()                                  ' 傳送: Read Data
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 5                                    ' Packet Length
            TX_Packet(3) = CMD_READ_DATA                        ' Command
            For i = 0 To 3
                sum += TX_Packet(i)
            Next
            TX_Packet(4) = sum And 255                          ' Checksum
            TX_CMD = CMD_READ_DATA
            TX_CTR = 0
            TX_Flag = True
            READ_Flag = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Black, 5, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 5, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 5)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (Read Data), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_SYNC_EN()                                    ' 傳送: SYNC Enable
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_SYNC_EN                          ' Command
            TX_Packet(4) = SYNC_EN
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_SYNC_EN
            TX_CTR = 0
            TX_Flag = True
            SYNC_EN_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Black, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (SYNC Enable), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L1_Set()                                     ' 傳送: L1 Set
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L1_SET                           ' Command
            TX_Packet(4) = L1_Set                               ' L1 設定值
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L1_SET
            TX_CTR = 0
            TX_Flag = True
            L1_Set_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Maroon, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L1 Set), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L2_Set()                                     ' 傳送: L2 Set
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L2_SET                           ' Command
            TX_Packet(4) = L2_Set                               ' L2 設定值
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L2_SET
            TX_CTR = 0
            TX_Flag = True
            L2_Set_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Maroon, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L2 Set), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L3_Set()                                     ' 傳送: L3 Set
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L3_SET                           ' Command
            TX_Packet(4) = L3_Set                               ' L3 設定值
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L3_SET
            TX_CTR = 0
            TX_Flag = True
            L3_Set_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Maroon, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L3 Set), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L4_Set()                                     ' 傳送: L4 Set
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L4_SET                           ' Command
            TX_Packet(4) = L4_Set                               ' L4 設定值
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L4_SET
            TX_CTR = 0
            TX_Flag = True
            L4_Set_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Maroon, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L4 Set), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L1_Output()                                  ' 傳送: L1 Output
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 7                                    ' Packet Length
            TX_Packet(3) = CMD_L1_OUTPUT                        ' Command
            TX_Packet(4) = L1_Output >> 8                       ' L1 Output, HB
            TX_Packet(5) = L1_Output And 255                    ' L1 Output, LB
            For i = 0 To 5
                sum += TX_Packet(i)
            Next
            TX_Packet(6) = sum And 255                          ' Checksum
            TX_CMD = CMD_L1_OUTPUT
            TX_CTR = 0
            TX_Flag = True
            L1_Output_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Blue, 7, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 7, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 7)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L1 Output), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L2_Output()                                  ' 傳送: L2 Output
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 7                                    ' Packet Length
            TX_Packet(3) = CMD_L2_OUTPUT                        ' Command
            TX_Packet(4) = L2_Output >> 8                       ' L2 Output, HB
            TX_Packet(5) = L2_Output And 255                    ' L2 Output, LB
            For i = 0 To 5
                sum += TX_Packet(i)
            Next
            TX_Packet(6) = sum And 255                          ' Checksum
            TX_CMD = CMD_L2_OUTPUT
            TX_CTR = 0
            TX_Flag = True
            L2_Output_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Blue, 7, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 7, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 7)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L2 Output), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L3_Output()                                  ' 傳送: L3 Output
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 7                                    ' Packet Length
            TX_Packet(3) = CMD_L3_OUTPUT                        ' Command
            TX_Packet(4) = L3_Output >> 8                       ' L3 Output, HB
            TX_Packet(5) = L3_Output And 255                    ' L3 Output, LB
            For i = 0 To 5
                sum += TX_Packet(i)
            Next
            TX_Packet(6) = sum And 255                          ' Checksum
            TX_CMD = CMD_L3_OUTPUT
            TX_CTR = 0
            TX_Flag = True
            L3_Output_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Blue, 7, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 7, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 7)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L3 Output), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L4_Output()                                  ' 傳送: L4 Output
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 7                                    ' Packet Length
            TX_Packet(3) = CMD_L4_OUTPUT                        ' Command
            TX_Packet(4) = L4_Output >> 8                       ' L4 Output, HB
            TX_Packet(5) = L4_Output And 255                    ' L4 Output, LB
            For i = 0 To 5
                sum += TX_Packet(i)
            Next
            TX_Packet(6) = sum And 255                          ' Checksum
            TX_CMD = CMD_L4_OUTPUT
            TX_CTR = 0
            TX_Flag = True
            L4_Output_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Blue, 7, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 7, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 7)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L4 Output), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L1_ON_OFF()                                  ' 傳送: L1 ON/OFF
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L1_ON_OFF                        ' Command
            TX_Packet(4) = L1_Set And SET_ON_OFF                ' L! ON/OFF
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L1_ON_OFF
            TX_CTR = 0
            TX_Flag = True
            L1_ON_OFF_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Green, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L1 ON/OFF), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L2_ON_OFF()                                  ' 傳送: L2 ON/OFF
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L2_ON_OFF                        ' Command
            TX_Packet(4) = L2_Set And SET_ON_OFF                ' L2 ON/OFF
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L2_ON_OFF
            TX_CTR = 0
            TX_Flag = True
            L2_ON_OFF_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Green, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L2 ON/OFF), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L3_ON_OFF()                                  ' 傳送: L3 ON/OFF
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L3_ON_OFF                        ' Command
            TX_Packet(4) = L3_Set And SET_ON_OFF                ' L3 ON/OFF
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L3_ON_OFF
            TX_CTR = 0
            TX_Flag = True
            L3_ON_OFF_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Green, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L3 ON/OFF), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    Private Sub TX_L4_ON_OFF()                                  ' 傳送: L4 ON/OFF
        Dim i As Byte
        Dim sum As UShort = 0

        If TX_Flag = False Then
            TX_Packet(0) = &H4D                                 ' Header 'M'
            TX_Packet(1) = &H53                                 ' Header 'S'
            TX_Packet(2) = 6                                    ' Packet Length
            TX_Packet(3) = CMD_L4_ON_OFF                        ' Command
            TX_Packet(4) = L4_Set And SET_ON_OFF                ' L4 ON/OFF
            For i = 0 To 4
                sum += TX_Packet(i)
            Next
            TX_Packet(5) = sum And 255                          ' Checksum
            TX_CMD = CMD_L4_ON_OFF
            TX_CTR = 0
            TX_Flag = True
            L4_ON_OFF_CHG = False
        End If

        TX_CTR += 1
        If TX_CTR = 1 Then
            Dsp_TX_Packet(Color.Green, 6, TX_CTR)
        Else
            Dsp_TX_Packet(Color.Magenta, 6, TX_CTR)
        End If
        ACK_TO_CTR = 0
        Auto_CTR = 0

        Try
            serialPort.Write(TX_Packet, 0, 6)
        Catch ex As Exception
            Dsp_String(Color.Red, " 傳送封包 (L4 ON/OFF), 發生不可預期的錯誤 !" + vbCrLf)
            SystemSounds.Beep.Play()
            COMPort_Disconnect()                                ' 串列埠離線
        End Try
    End Sub

    ' 串列埠接收 ********************************************************************

    Private Sub serialPort_DataReceived(sender As Object, e As SerialDataReceivedEventArgs) Handles serialPort.DataReceived
        Dim bytes As Integer = serialPort.BytesToRead

        If bytes > 0 Then
            serialPort.Read(RX_Buffer, 0, bytes)
            Packet_Check(bytes)                                 ' 封包資料判斷
        End If
    End Sub

    Private Sub Packet_Check(ByVal length As Integer)           ' 封包資料判斷
        Dim i, n As Byte
        Dim sum As UShort

        For i = 0 To (length - 1)
            RX_Packet(RX_Index) = RX_Buffer(i)
            If RX_Index = 0 Then                                ' Header 'S'
                If RX_Packet(0) = &H53 Then
                    RX_Index += 1
                End If
            Else
                If RX_Index = 1 Then                            ' Header 'M'
                    If RX_Packet(1) = &H4D Then
                        RX_Index += 1
                    Else
                        RX_Index = 0
                    End If
                Else
                    If RX_Index = 2 Then                        ' Packet Length
                        If (RX_Packet(2) = 17) Or (RX_Packet(2) = 5) Then
                            RX_Index += 1
                            RX_Length = RX_Packet(2)
                        Else
                            RX_Index = 0
                        End If
                    Else
                        RX_Index += 1
                        If RX_Index >= RX_Length Then
                            RX_Index = 0
                            sum = 0
                            For n = 0 To (RX_Length - 2)
                                sum += RX_Packet(n)
                            Next
                            If (sum And 255) = RX_Packet(n) Then
                                Packet_OK()                     ' 封包資料格式 OK
                            End If
                        End If
                    End If
                End If
            End If
        Next
    End Sub

    Private Sub Packet_OK()                                     ' 封包資料格式 OK
        Select Case TX_CMD
            Case CMD_READ_DATA
                RX_Read_Data()                                  ' 接收: Read Data
            Case CMD_SYNC_EN
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L1_SET
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L2_SET
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L3_SET
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L4_SET
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L1_OUTPUT
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L2_OUTPUT
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L3_OUTPUT
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L4_OUTPUT
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L1_ON_OFF
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L2_ON_OFF
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L3_ON_OFF
                RX_ACK()                                        ' 接收: ACK
            Case CMD_L4_ON_OFF
                RX_ACK()                                        ' 接收: ACK
        End Select
    End Sub

    Private Sub RX_Read_Data()                                  ' 接收: Read Data
        If RX_Length = 17 Then
            RX_Flag = True
        End If
    End Sub

    Private Sub RX_ACK()                                        ' 接收: ACK
        If RX_Length = 5 Then
            RX_Flag = True
        End If
    End Sub

    ' L1輸出調整 ***************************************************************

    Private Sub cbb_L1_MaxC_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cbb_L1_MaxC.SelectedIndexChanged
        If ((L1_Set >> 4) And &H3) <> cbb_L1_MaxC.SelectedIndex Then
            Dim v As Byte
            v = L1_Set And (SET_EXT_CTRL Or SET_ON_OFF)
            L1_Set = (cbb_L1_MaxC.SelectedIndex << 4) Or v

            L1_Output = 0
            nud_L1_Output.Value = 0
            tkb_L1_Output.Value = 0
            nud_L1_Output.Maximum = MaxCurrent(cbb_L1_MaxC.SelectedIndex)
            tkb_L1_Output.Maximum = nud_L1_Output.Maximum
            tkb_L1_Output.TickFrequency = tkb_L1_Output.Maximum / 10
            L1_Output_CHG = False

            If READ_Flag = True Then
                L1_Set_CHG = True
            End If
        End If
    End Sub

    Private Sub nud_L1_Output_ValueChanged(sender As Object, e As EventArgs) Handles nud_L1_Output.ValueChanged
        Dim MaxC As Integer

        tkb_L1_Output.Value = nud_L1_Output.Value
        L1_Output = tkb_L1_Output.Value
        MaxC = MaxCurrent(cbb_L1_MaxC.SelectedIndex)
        lb_L1_PCV.Text = String.Format("{0:F1}", (L1_Output * 100) / MaxC)

        If READ_Flag = True Then
            L1_Output_CHG = True
        End If
    End Sub

    Private Sub tkb_L1_Output_ValueChanged(sender As Object, e As EventArgs) Handles tkb_L1_Output.ValueChanged
        Dim MaxC As Integer

        nud_L1_Output.Value = tkb_L1_Output.Value
        L1_Output = tkb_L1_Output.Value
        MaxC = MaxCurrent(cbb_L1_MaxC.SelectedIndex)
        lb_L1_PCV.Text = String.Format("{0:F1}", (L1_Output * 100) / MaxC)

        If READ_Flag = True Then
            L1_Output_CHG = True
        End If
    End Sub

    ' L1 ON/OFF ================================================================

    Private Sub ckb_L1_EXT_CheckedChanged(sender As Object, e As EventArgs) Handles ckb_L1_EXT.CheckedChanged
        If ckb_L1_EXT.Checked = False Then                      ' Enable to Disable
            pn_L1_LED.Visible = True
            btn_L1.Visible = True
            L1_Set = L1_Set And (Not SET_EXT_CTRL)              ' L1外部ON/OFF控制 = Disable
        Else                                                    ' Disable to Enable
            pn_L1_LED.Visible = False
            btn_L1.Visible = False
            L1_Set = L1_Set Or SET_EXT_CTRL                     ' L1外部ON/OFF控制 = Enable
        End If

        If READ_Flag = True Then
            L1_Set_CHG = True
        End If

        Select Case cbb_SYNC_EN.SelectedIndex
            Case SYNC2
                ckb_L2_EXT.Checked = ckb_L1_EXT.Checked
            Case SYNC3
                ckb_L2_EXT.Checked = ckb_L1_EXT.Checked
                ckb_L3_EXT.Checked = ckb_L1_EXT.Checked
            Case SYNC4
                ckb_L2_EXT.Checked = ckb_L1_EXT.Checked
                ckb_L3_EXT.Checked = ckb_L1_EXT.Checked
                ckb_L4_EXT.Checked = ckb_L1_EXT.Checked
        End Select
    End Sub

    Private Sub btn_L1_Click(sender As Object, e As EventArgs) Handles btn_L1.Click
        If (L1_Set And SET_ON_OFF) = 0 Then                     ' OFF to ON
            pn_L1_LED.BackColor = Color.Green
            btn_L1.Text = "OFF"
            L1_Set = L1_Set Or SET_ON_OFF                       ' L1 ON/OFF = ON
        Else                                                    ' ON to OFF
            pn_L1_LED.BackColor = Color.Black
            btn_L1.Text = "ON"
            L1_Set = L1_Set And (Not SET_ON_OFF)                ' L1 ON/OFF = OFF
        End If

        If READ_Flag = True Then
            L1_ON_OFF_CHG = True
        End If

        If SYNC_EN > ASYNC Then
            pn_L2_LED.BackColor = pn_L1_LED.BackColor
            btn_L2.Text = btn_L1.Text
            If (L1_Set And SET_ON_OFF) = 0 Then
                L2_Set = L2_Set And (Not SET_ON_OFF)
            Else
                L2_Set = L2_Set Or SET_ON_OFF
            End If
        End If

        If SYNC_EN > SYNC2 Then
            pn_L3_LED.BackColor = pn_L1_LED.BackColor
            btn_L3.Text = btn_L1.Text
            If (L1_Set And SET_ON_OFF) = 0 Then
                L3_Set = L3_Set And (Not SET_ON_OFF)
            Else
                L3_Set = L3_Set Or SET_ON_OFF
            End If
        End If

        If SYNC_EN > SYNC3 Then
            pn_L4_LED.BackColor = pn_L1_LED.BackColor
            btn_L4.Text = btn_L1.Text
            If (L1_Set And SET_ON_OFF) = 0 Then
                L4_Set = L4_Set And (Not SET_ON_OFF)
            Else
                L4_Set = L4_Set Or SET_ON_OFF
            End If
        End If
    End Sub

    ' L2輸出調整 ***************************************************************

    Private Sub cbb_L2_MaxC_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cbb_L2_MaxC.SelectedIndexChanged
        If ((L2_Set >> 4) And &H3) <> cbb_L2_MaxC.SelectedIndex Then
            Dim v As Byte
            v = L2_Set And (SET_EXT_CTRL Or SET_ON_OFF)
            L2_Set = (cbb_L2_MaxC.SelectedIndex << 4) Or v

            L2_Output = 0
            nud_L2_Output.Value = 0
            tkb_L2_Output.Value = 0
            nud_L2_Output.Maximum = MaxCurrent(cbb_L2_MaxC.SelectedIndex)
            tkb_L2_Output.Maximum = nud_L2_Output.Maximum
            tkb_L2_Output.TickFrequency = tkb_L2_Output.Maximum / 10
            L2_Output_CHG = False

            If READ_Flag = True Then
                L2_Set_CHG = True
            End If
        End If
    End Sub

    Private Sub nud_L2_Output_ValueChanged(sender As Object, e As EventArgs) Handles nud_L2_Output.ValueChanged
        Dim MaxC As Integer

        tkb_L2_Output.Value = nud_L2_Output.Value
        L2_Output = tkb_L2_Output.Value
        MaxC = MaxCurrent(cbb_L2_MaxC.SelectedIndex)
        lb_L2_PCV.Text = String.Format("{0:F1}", (L2_Output * 100) / MaxC)

        If READ_Flag = True Then
            L2_Output_CHG = True
        End If
    End Sub

    Private Sub tkb_L2_Output_ValueChanged(sender As Object, e As EventArgs) Handles tkb_L2_Output.ValueChanged
        Dim MaxC As Integer

        nud_L2_Output.Value = tkb_L2_Output.Value
        L2_Output = tkb_L2_Output.Value
        MaxC = MaxCurrent(cbb_L2_MaxC.SelectedIndex)
        lb_L2_PCV.Text = String.Format("{0:F1}", (L2_Output * 100) / MaxC)

        If READ_Flag = True Then
            L2_Output_CHG = True
        End If
    End Sub

    ' L2 ON/OFF ================================================================

    Private Sub ckb_L2_EXT_CheckedChanged(sender As Object, e As EventArgs) Handles ckb_L2_EXT.CheckedChanged
        If ckb_L2_EXT.Checked = False Then                      ' Enable to Disable
            pn_L2_LED.Visible = True
            btn_L2.Visible = True
            L2_Set = L2_Set And (Not SET_EXT_CTRL)              ' L2外部ON/OFF控制 = Disable
        Else                                                    ' Disable to Enable
            pn_L2_LED.Visible = False
            btn_L2.Visible = False
            L2_Set = L2_Set Or SET_EXT_CTRL                     ' L2外部ON/OFF控制 = Enable
        End If

        If (READ_Flag = True) And (SYNC_EN < SYNC2) Then
            L2_Set_CHG = True
        End If
    End Sub

    Private Sub btn_L2_Click(sender As Object, e As EventArgs) Handles btn_L2.Click
        If (L2_Set And SET_ON_OFF) = 0 Then                     ' OFF to ON
            pn_L2_LED.BackColor = Color.Green
            btn_L2.Text = "OFF"
            L2_Set = L2_Set Or SET_ON_OFF                       ' L2 ON/OFF = ON
        Else                                                    ' ON to OFF
            pn_L2_LED.BackColor = Color.Black
            btn_L2.Text = "ON"
            L2_Set = L2_Set And (Not SET_ON_OFF)                ' L2 ON/OFF = OFF
        End If

        If (READ_Flag = True) And (SYNC_EN < SYNC2) Then
            L2_ON_OFF_CHG = True
        End If
    End Sub

    ' L3輸出調整 ***************************************************************

    Private Sub cbb_L3_MaxC_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cbb_L3_MaxC.SelectedIndexChanged
        If ((L3_Set >> 4) And &H3) <> cbb_L3_MaxC.SelectedIndex Then
            Dim v As Byte
            v = L3_Set And (SET_EXT_CTRL Or SET_ON_OFF)
            L3_Set = (cbb_L3_MaxC.SelectedIndex << 4) Or v

            L3_Output = 0
            nud_L3_Output.Value = 0
            tkb_L3_Output.Value = 0
            nud_L3_Output.Maximum = MaxCurrent(cbb_L3_MaxC.SelectedIndex)
            tkb_L3_Output.Maximum = nud_L3_Output.Maximum
            tkb_L3_Output.TickFrequency = tkb_L3_Output.Maximum / 10
            L3_Output_CHG = False

            If READ_Flag = True Then
                L3_Set_CHG = True
            End If
        End If
    End Sub

    Private Sub nud_L3_Output_ValueChanged(sender As Object, e As EventArgs) Handles nud_L3_Output.ValueChanged
        Dim MaxC As Integer

        tkb_L3_Output.Value = nud_L3_Output.Value
        L3_Output = tkb_L3_Output.Value
        MaxC = MaxCurrent(cbb_L3_MaxC.SelectedIndex)
        lb_L3_PCV.Text = String.Format("{0:F1}", (L3_Output * 100) / MaxC)

        If READ_Flag = True Then
            L3_Output_CHG = True
        End If
    End Sub

    Private Sub tkb_L3_Output_ValueChanged(sender As Object, e As EventArgs) Handles tkb_L3_Output.ValueChanged
        Dim MaxC As Integer

        nud_L3_Output.Value = tkb_L3_Output.Value
        L3_Output = tkb_L3_Output.Value
        MaxC = MaxCurrent(cbb_L3_MaxC.SelectedIndex)
        lb_L3_PCV.Text = String.Format("{0:F1}", (L3_Output * 100) / MaxC)

        If READ_Flag = True Then
            L3_Output_CHG = True
        End If
    End Sub

    ' L3 ON/OFF ================================================================

    Private Sub ckb_L3_EXT_CheckedChanged(sender As Object, e As EventArgs) Handles ckb_L3_EXT.CheckedChanged
        If ckb_L3_EXT.Checked = False Then                      ' Enable to Disable
            pn_L3_LED.Visible = True
            btn_L3.Visible = True
            L3_Set = L3_Set And (Not SET_EXT_CTRL)              ' L3外部ON/OFF控制 = Disable
        Else                                                    ' Disable to Enable
            pn_L3_LED.Visible = False
            btn_L3.Visible = False
            L3_Set = L3_Set Or SET_EXT_CTRL                     ' L3外部ON/OFF控制 = Enable
        End If

        If (READ_Flag = True) And (SYNC_EN < SYNC3) Then
            L3_Set_CHG = True
        End If
    End Sub

    Private Sub btn_L3_Click(sender As Object, e As EventArgs) Handles btn_L3.Click
        If (L3_Set And SET_ON_OFF) = 0 Then                     ' OFF to ON
            pn_L3_LED.BackColor = Color.Green
            btn_L3.Text = "OFF"
            L3_Set = L3_Set Or SET_ON_OFF                       ' L3 ON/OFF = ON
        Else                                                    ' ON to OFF
            pn_L3_LED.BackColor = Color.Black
            btn_L3.Text = "ON"
            L3_Set = L3_Set And (Not SET_ON_OFF)                ' L3 ON/OFF = OFF
        End If

        If (READ_Flag = True) And (SYNC_EN < SYNC3) Then
            L3_ON_OFF_CHG = True
        End If
    End Sub

    ' L4輸出調整 ***************************************************************

    Private Sub cbb_L4_MaxC_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cbb_L4_MaxC.SelectedIndexChanged
        If ((L4_Set >> 4) And &H3) <> cbb_L4_MaxC.SelectedIndex Then
            Dim v As Byte
            v = L4_Set And (SET_EXT_CTRL Or SET_ON_OFF)
            L4_Set = (cbb_L4_MaxC.SelectedIndex << 4) Or v

            L4_Output = 0
            nud_L4_Output.Value = 0
            tkb_L4_Output.Value = 0
            nud_L4_Output.Maximum = MaxCurrent(cbb_L4_MaxC.SelectedIndex)
            tkb_L4_Output.Maximum = nud_L4_Output.Maximum
            tkb_L4_Output.TickFrequency = tkb_L4_Output.Maximum / 10
            L4_Output_CHG = False

            If READ_Flag = True Then
                L4_Set_CHG = True
            End If
        End If
    End Sub

    Private Sub nud_L4_Output_ValueChanged(sender As Object, e As EventArgs) Handles nud_L4_Output.ValueChanged
        Dim MaxC As Integer

        tkb_L4_Output.Value = nud_L4_Output.Value
        L4_Output = tkb_L4_Output.Value
        MaxC = MaxCurrent(cbb_L4_MaxC.SelectedIndex)
        lb_L4_PCV.Text = String.Format("{0:F1}", (L4_Output * 100) / MaxC)

        If READ_Flag = True Then
            L4_Output_CHG = True
        End If
    End Sub

    Private Sub tkb_L4_Output_ValueChanged(sender As Object, e As EventArgs) Handles tkb_L4_Output.ValueChanged
        Dim MaxC As Integer

        nud_L4_Output.Value = tkb_L4_Output.Value
        L4_Output = tkb_L4_Output.Value
        MaxC = MaxCurrent(cbb_L4_MaxC.SelectedIndex)
        lb_L4_PCV.Text = String.Format("{0:F1}", (L4_Output * 100) / MaxC)

        If READ_Flag = True Then
            L4_Output_CHG = True
        End If
    End Sub

    ' L4 ON/OFF ================================================================

    Private Sub ckb_L4_EXT_CheckedChanged(sender As Object, e As EventArgs) Handles ckb_L4_EXT.CheckedChanged
        If ckb_L4_EXT.Checked = False Then                      ' Enable to Disable
            pn_L4_LED.Visible = True
            btn_L4.Visible = True
            L4_Set = L4_Set And (Not SET_EXT_CTRL)              ' L4外部ON/OFF控制 = Disable
        Else                                                    ' Disable to Enable
            pn_L4_LED.Visible = False
            btn_L4.Visible = False
            L4_Set = L4_Set Or SET_EXT_CTRL                     ' L4外部ON/OFF控制 = Enable
        End If

        If (READ_Flag = True) And (SYNC_EN <> SYNC4) Then
            L4_Set_CHG = True
        End If
    End Sub

    Private Sub btn_L4_Click(sender As Object, e As EventArgs) Handles btn_L4.Click
        If (L4_Set And SET_ON_OFF) = 0 Then                     ' OFF to ON
            pn_L4_LED.BackColor = Color.Green
            btn_L4.Text = "OFF"
            L4_Set = L4_Set Or SET_ON_OFF                       ' L4 ON/OFF = ON
        Else                                                    ' ON to OFF
            pn_L4_LED.BackColor = Color.Black
            btn_L4.Text = "ON"
            L4_Set = L4_Set And (Not SET_ON_OFF)                ' L4 ON/OFF = OFF
        End If

        If (READ_Flag = True) And (SYNC_EN <> SYNC4) Then
            L4_ON_OFF_CHG = True
        End If
    End Sub

    ' SYNC Enable **************************************************************

    Private Sub cbb_SYNC_EN_SelectedIndexChanged(sender As Object, e As EventArgs) Handles cbb_SYNC_EN.SelectedIndexChanged
        SystemSounds.Exclamation.Play()
        SYNC_EN = cbb_SYNC_EN.SelectedIndex
        Select Case SYNC_EN
            Case ASYNC
                If (ACK_State And AS_L2_ERR) = 0 Then
                    ckb_L2_EXT.Enabled = True
                    btn_L2.Enabled = True
                End If
                If (ACK_State And AS_L3_ERR) = 0 Then
                    ckb_L3_EXT.Enabled = True
                    btn_L3.Enabled = True
                End If
                If (ACK_State And AS_L4_ERR) = 0 Then
                    ckb_L4_EXT.Enabled = True
                    btn_L4.Enabled = True
                End If
            Case SYNC2
                ckb_L2_EXT.Enabled = False
                btn_L2.Enabled = False
                If (ACK_State And AS_L3_ERR) = 0 Then
                    ckb_L3_EXT.Enabled = True
                    btn_L3.Enabled = True
                End If
                If (ACK_State And AS_L4_ERR) = 0 Then
                    ckb_L4_EXT.Enabled = True
                    btn_L4.Enabled = True
                End If
            Case SYNC3
                ckb_L2_EXT.Enabled = False
                btn_L2.Enabled = False
                ckb_L3_EXT.Enabled = False
                btn_L3.Enabled = False
                If (ACK_State And AS_L4_ERR) = 0 Then
                    ckb_L4_EXT.Enabled = True
                    btn_L4.Enabled = True
                End If
            Case SYNC4
                ckb_L2_EXT.Enabled = False
                btn_L2.Enabled = False
                ckb_L3_EXT.Enabled = False
                btn_L3.Enabled = False
                ckb_L4_EXT.Enabled = False
                btn_L4.Enabled = False
        End Select

        If SYNC_EN > ASYNC Then
            SYNC_L2()                                           ' ON/OFF同步控制: L2
        End If
        If SYNC_EN > SYNC2 Then
            SYNC_L3()                                           ' ON/OFF同步控制: L3
        End If
        If SYNC_EN > SYNC3 Then
            SYNC_L4()                                           ' ON/OFF同步控制: L4
        End If

        If READ_Flag = True Then
            SYNC_EN_CHG = True
        End If
    End Sub

    Private Sub SYNC_L2()                                       ' ON/OFF同步控制: L2
        ckb_L2_EXT.Checked = ckb_L1_EXT.Checked
        pn_L2_LED.BackColor = pn_L1_LED.BackColor
        pn_L2_LED.Visible = pn_L1_LED.Visible
        btn_L2.Text = btn_L1.Text
        btn_L2.Visible = btn_L1.Visible

        If (L1_Set And SET_EXT_CTRL) = 0 Then
            L2_Set = L2_Set And (Not SET_EXT_CTRL)
        Else
            L2_Set = L2_Set Or SET_EXT_CTRL
        End If

        If (L1_Set And SET_ON_OFF) = 0 Then
            L2_Set = L2_Set And (Not SET_ON_OFF)
        Else
            L2_Set = L2_Set Or SET_ON_OFF
        End If
    End Sub

    Private Sub SYNC_L3()                                       ' ON/OFF同步控制: L3
        ckb_L3_EXT.Checked = ckb_L1_EXT.Checked
        pn_L3_LED.BackColor = pn_L1_LED.BackColor
        pn_L3_LED.Visible = pn_L1_LED.Visible
        btn_L3.Text = btn_L1.Text
        btn_L3.Visible = btn_L1.Visible

        If (L1_Set And SET_EXT_CTRL) = 0 Then
            L3_Set = L3_Set And (Not SET_EXT_CTRL)
        Else
            L3_Set = L3_Set Or SET_EXT_CTRL
        End If

        If (L1_Set And SET_ON_OFF) = 0 Then
            L3_Set = L3_Set And (Not SET_ON_OFF)
        Else
            L3_Set = L3_Set Or SET_ON_OFF
        End If
    End Sub

    Private Sub SYNC_L4()                                       ' ON/OFF同步控制: L4
        ckb_L4_EXT.Checked = ckb_L1_EXT.Checked
        pn_L4_LED.BackColor = pn_L1_LED.BackColor
        pn_L4_LED.Visible = pn_L1_LED.Visible
        btn_L4.Text = btn_L1.Text
        btn_L4.Visible = btn_L1.Visible

        If (L1_Set And SET_EXT_CTRL) = 0 Then
            L4_Set = L4_Set And (Not SET_EXT_CTRL)
        Else
            L4_Set = L4_Set Or SET_EXT_CTRL
        End If

        If (L1_Set And SET_ON_OFF) = 0 Then
            L4_Set = L4_Set And (Not SET_ON_OFF)
        Else
            L4_Set = L4_Set Or SET_ON_OFF
        End If
    End Sub
End Class
